@@ -1,50 +1,63 @@
 from __future__ import annotations
 
-import shutil
 from io import BytesIO
 from typing import Any
 
 from bailo.core.agent import Agent, TokenAgent
-from bailo.core.enums import ModelVisibility, SchemaKind
+from bailo.core.enums import CollaboratorEntry, EntryKind, ModelVisibility, SchemaKind
 from bailo.core.utils import filter_none
 
 
 class Client:
-    """
-    Creates a Client object that can be used to talk to the website.
+    """Create a Client object that can be used to talk to the website.
 
     :param url: Url of bailo website
-    :param agent:
+    :param agent: An agent object to handle requests
     """
 
     def __init__(self, url: str, agent: Agent = Agent()):
+        """Initialise a Client.
+
+        :param url: URL of the Bailo instance website.
+        :param agent: An agent object to handle requests, defaults to Agent().
+        """
         self.url = url.rstrip("/") + "/api"
         self.agent = agent
 
     def post_model(
         self,
         name: str,
+        kind: EntryKind,
         description: str,
-        team_id: str,
         visibility: ModelVisibility | None = None,
+        organisation: str | None = None,
+        state: str | None = None,
+        collaborators: list[CollaboratorEntry] | None = None,
     ):
-        """
-        Creates a model.
+        """Create a model.
 
         :param name: Name of the model
+        :param kind: Either a Model or a Datacard
         :param description: Description of the model
-        :param visibility: Enum to define model visibility (e.g public or private)
+        :param visibility: Enum to define model visibility (e.g public or private), defaults to None
+        :param organisation: Organisation responsible for the model, defaults to None
+        :param state: Development readiness of the model, defaults to None
+        :param collaborators: list of CollaboratorEntry to define who the model's collaborators (a.k.a. model access) are, defaults to None
         :return: JSON response object
         """
+        _visibility: str = "public"
         if visibility is not None:
-            visibility = str(visibility)
+            _visibility = str(visibility)
 
         filtered_json = filter_none(
             {
                 "name": name,
+                "kind": kind,
                 "description": description,
-                "visibility": visibility,
-                "teamId": team_id,
+                "visibility": _visibility,
+                "organisation": organisation,
+                "state": state,
+                "collaborators": collaborators,
             }
         )
 
@@ -63,8 +76,8 @@ class Client:
         """Find and returns a list of models based on provided search terms.
 
         :param task: Model task (e.g. image classification), defaults to None
-        :param libraries: Model library (e.g. TensorFlow), defaults to []
-        :param filters: Custom filters, defaults to []
+        :param libraries: Model library (e.g. TensorFlow), defaults to None
+        :param filters: Custom filters, defaults to None
         :param search: String to be located in model cards, defaults to ""
         :return: JSON response object
         """
@@ -88,8 +101,7 @@ class Client:
         self,
         model_id: str,
     ):
-        """
-        Retrieves a specific model using its unique ID.
+        """Retrieve a specific model using its unique ID.
 
         :param model_id: Unique model ID
         :return: JSON response object
@@ -102,19 +114,36 @@ class Client:
         self,
         model_id: str,
         name: str | None = None,
+        kind: str | None = None,
         description: str | None = None,
         visibility: str | None = None,
+        organisation: str | None = None,
+        state: str | None = None,
+        collaborators: list[CollaboratorEntry] | None = None,
     ):
-        """
-        Updates a specific model using its unique ID.
+        """Update a specific model using its unique ID.
 
         :param model_id: Unique model ID
         :param name: Name of the model, defaults to None
+        :param kind: Either a Model or a Datacard, defaults to None
         :param description: Description of the model, defaults to None
-        :param visibility: Enum to define model visibility (e.g. public or private), defaults to None
+        :param visibility: Enum to define model visibility (e.g public or private), defaults to None
+        :param organisation: Organisation responsible for the model, defaults to None
+        :param state: Development readiness of the model, defaults to None
+        :param collaborators: list of CollaboratorEntry to define who the model's collaborators (a.k.a. model access) are, defaults to None
         :return: JSON response object
         """
-        filtered_json = filter_none({"name": name, "description": description, "visibility": visibility})
+        filtered_json = filter_none(
+            {
+                "name": name,
+                "organisation": organisation,
+                "state": state,
+                "kind": kind,
+                "description": description,
+                "visibility": visibility,
+                "collaborators": collaborators,
+            }
+        )
 
         return self.agent.patch(f"{self.url}/v2/model/{model_id}", json=filtered_json).json()
 
@@ -123,8 +152,7 @@ class Client:
         model_id: str,
         version: str,
     ):
-        """
-        Retrieves a specific model card, using the unique model ID and version.
+        """Retrieve a specific model card, using the unique model ID and version.
 
         :param model_id: Unique model ID
         :param version: Model card version
@@ -139,8 +167,7 @@ class Client:
         model_id: str,
         metadata: Any,
     ):
-        """
-        Updates the latest model card, using the unique model ID.
+        """Update the latest model card, using the unique model ID.
 
         :param model_id: Unique model ID
         :param metadata: Metadata object, defined by model card schema
@@ -159,7 +186,7 @@ class Client:
         schema_id: str,
     ):
         """
-        Creates a model card using a given schema ID.
+        Create a model card using a given schema ID.
 
         :param model_id: Unique model ID
         :param schema_id: Unique model card schema ID
@@ -172,6 +199,17 @@ class Client:
             },
         ).json()
 
+    def model_card_from_template(self, model_id: str, template_id: str | None):
+        """Create a model card using a given template ID (previously created models, model ID)
+        :param model_id: Unique model ID
+        :param template_id Previous model's unique ID to be used as template for new model card
+        :return: JSON response object
+        """
+        return self.agent.post(
+            f"{self.url}/v2/model/{model_id}/setup/from-template",
+            json={"templateId": template_id},
+        ).json()
+
     def post_release(
         self,
         model_id: str,
@@ -184,7 +222,7 @@ class Client:
         draft: bool | None = False,
     ):
         """
-        Creates a new model release.
+        Create a new model release.
 
         :param model_id: Unique model ID
         :param model_card_version: Model card version
@@ -212,6 +250,7 @@ class Client:
     def put_release(
         self,
         model_id: str,
+        model_card_version: int,
         release_version: str,
         notes: str,
         draft: bool,
@@ -219,7 +258,7 @@ class Client:
         images: list[str],
     ):
         """
-        Creates a new model release.
+        Create a new model release.
 
         :param model_id: Unique model ID
         :param model_card_version: Model card version
@@ -238,6 +277,7 @@ class Client:
                 "draft": draft,
                 "fileIds": file_ids,
                 "images": images,
+                "modelCardVersion": model_card_version,
             },
         ).json()
 
@@ -246,7 +286,7 @@ class Client:
         model_id: str,
     ):
         """
-        Gets all releases for a model.
+        Get all releases for a model.
 
         :param model_id: Unique model ID
         :return: JSON response object
@@ -257,7 +297,7 @@ class Client:
 
     def get_release(self, model_id: str, release_version: str):
         """
-        Gets a specific model release.
+        Get a specific model release.
 
         :param model_id: Unique model ID
         :param release_version: Release version
@@ -273,7 +313,7 @@ class Client:
         release_version: str,
     ):
         """
-        Deletes a specific model release.
+        Delete a specific model release.
 
         :param model_id: Unique model ID
         :param release_version: Release version
@@ -288,7 +328,7 @@ class Client:
         model_id: str,
     ):
         """
-        Gets files for a model.
+        Get files for a model.
 
         :param model_id: Unique model ID
         :return: JSON response object
@@ -301,33 +341,59 @@ class Client:
         self,
         model_id: str,
         file_id: str,
-        buffer: BytesIO,
     ):
-        """
-        Downloads a specific file.
+        """Download a specific file by it's id.
 
         :param model_id: Unique model ID
         :param file_id: Unique file ID
-        :param buffer: BytesIO object for bailo to write to
         :return: The unique file ID
         """
         if isinstance(self.agent, TokenAgent):
-            req = self.agent.get(
-                f"{self.url}/v2/token/model/{model_id}/file/{file_id}/download", stream=True, timeout=10_000
+            return self.agent.get(
+                f"{self.url}/v2/token/model/{model_id}/file/{file_id}/download",
+                stream=True,
+                timeout=10_000,
             )
         else:
-            req = self.agent.get(f"{self.url}/v2/model/{model_id}/file/{file_id}/download", stream=True, timeout=10_000)
+            return self.agent.get(
+                f"{self.url}/v2/model/{model_id}/file/{file_id}/download",
+                stream=True,
+                timeout=10_000,
+            )
 
-        shutil.copyfileobj(req.raw, buffer)
-        return file_id
-
-    def simple_upload(self, model_id: str, name: str, buffer: BytesIO):
-        """
-        Creates a simple file upload.
+    def get_download_by_filename(
+        self,
+        model_id: str,
+        semver: str,
+        filename: str,
+    ):
+        """Download a specific file.
 
         :param model_id: Unique model ID
-        :param name: File name
-        :return: JSON response object
+        :param semver: Semver of the release
+        :param filename: The filename trying to download from
+        :return: The filename
+        """
+        if isinstance(self.agent, TokenAgent):
+            return self.agent.get(
+                f"{self.url}/v2/token/model/{model_id}/release/{semver}/file/{filename}/download",
+                stream=True,
+                timeout=10_000,
+            )
+        else:
+            return self.agent.get(
+                f"{self.url}/v2/model/{model_id}/release/{semver}/file/{filename}/download",
+                stream=True,
+                timeout=10_000,
+            )
+
+    def simple_upload(self, model_id: str, name: str, buffer: BytesIO):
+        """Upload a file associated with a model.
+
+        :param model_id: Unique model ID.
+        :param name: Name of the file to upload.
+        :param buffer: File-like BytesIO object containing the data to upload.
+        :return: Response object from the upload endpoint.
         """
         return self.agent.post(
             f"{self.url}/v2/model/{model_id}/files/upload/simple",
@@ -346,23 +412,21 @@ class Client:
         model_id: str,
         file_id: str,
     ):
-        """
-        Deletes a specific file associated with a model.
+        """Delete a specific file associated with a model.
 
         :param model_id: Unique model ID
         :param file_id: Unique file ID
         :return: JSON response object
         """
         return self.agent.delete(
-            f"{self.url}/v2/model/{model_id}/files/{file_id}",
+            f"{self.url}/v2/model/{model_id}/file/{file_id}",
         ).json()
 
     def get_all_images(
         self,
         model_id: str,
     ):
-        """
-        Gets all images.
+        """Get all images.
 
         :param model_id: A unique model ID
         :return: JSON response object
@@ -373,8 +437,7 @@ class Client:
         self,
         kind: SchemaKind | None = None,
     ):
-        """
-        Gets all schemas.
+        """Get all schemas.
 
         :param kind: Enum to define schema kind (e.g. Model or AccessRequest), defaults to None
         :return: JSON response object
@@ -388,8 +451,7 @@ class Client:
         self,
         schema_id: str,
     ):
-        """
-        Retrieves a specific schema using its unique ID.
+        """Retrieve a specific schema using its unique ID.
 
         :param schema_id: Unique schema ID
         :return: JSON response object.
@@ -405,15 +467,16 @@ class Client:
         description: str,
         kind: SchemaKind,
         json_schema: dict[str, Any],
+        review_roles: list[str],
     ):
-        """
-        Creates a schema.
+        """Create a schema.
 
         :param schema_id: Unique schema ID
         :param name: Name of the schema
         :param description: Description for the schema
         :param kind: Enum to define schema kind (e.g. Model or AccessRequest)
         :param json_schema: JSON schema
+        :param review_roles: list made up of the "shortName" property from a Review Role object
         :return: JSON response object
         """
         return self.agent.post(
@@ -424,6 +487,7 @@ class Client:
                 "description": description,
                 "kind": str(kind),
                 "jsonSchema": json_schema,
+                "reviewRoles": review_roles,
             },
         ).json()
 
@@ -433,38 +497,36 @@ class Client:
         model_id: str | None = None,
         version: str | None = None,
     ):
-        """
-        Gets all reviews within given parameters.
+        """Get all reviews within given parameters.
 
         :param active: Boolean representing status of review
         :param model_id: Unique model ID, defaults to None
         :param version: Model version, defaults to None
         :return: JSON response object.
         """
-        active = str(active).lower()
+        _active = str(active).lower()
 
         return self.agent.get(
             f"{self.url}/v2/reviews",
             params={
-                "active": active,
+                "active": _active,
                 "modelId": model_id,
                 "semver": version,
             },
         ).json()
 
-    def post_review(
+    def post_release_review(
         self,
         model_id: str,
+        version: str,
         role: str,
         decision: str,
-        version: str | None = None,
         comment: str | None = None,
     ):
-        """
-        Creates a review for a release
+        """Create a review for a release.
 
         :param model_id: A unique model ID
-        :param version: A semantic version for a release
+        :param version: Semver of the release
         :param role: The role of the user making the review
         :param decision: Either approve or request changes
         :param comment: A comment to go with the review
@@ -480,7 +542,7 @@ class Client:
         model_id: str,
     ):
         """
-        Gets roles for a model.
+        Get roles for a model.
 
         :param model_id: Unique model ID
         :return: JSON response object
@@ -489,105 +551,8 @@ class Client:
             f"{self.url}/v2/model/{model_id}/roles",
         ).json()
 
-    def get_model_user_roles(
-        self,
-        model_id: str,
-    ):
-        """
-        Gets current users roles for a model.
-
-        :param model_id: Unique model ID
-        :return: JSON response object
-        """
-        return self.agent.get(
-            f"{self.url}/v2/model/{model_id}/roles/mine",
-        ).json()
-
-    def post_team(
-        self,
-        team_id: str,
-        name: str,
-        description: str,
-    ):
-        """
-        Create new team.
-
-        :param team_id: Unique team ID
-        :param name: Team name
-        :param description: Team description
-        :return: JSON response object
-        """
-        return self.agent.post(
-            f"{self.url}/v2/teams",
-            json={
-                "id": team_id,
-                "name": name,
-                "description": description,
-            },
-        ).json()
-
-    def get_all_teams(
-        self,
-    ):
-        """
-        Get all teams.
-
-        :return: JSON response object
-        """
-        return self.agent.get(
-            f"{self.url}/v2/teams",
-        ).json()
-
-    def get_user_teams(
-        self,
-    ):
-        """
-        Get user teams.
-
-        :return: JSON response object
-        """
-        return self.agent.get(
-            f"{self.url}/v2/teams/mine",
-        ).json()
-
-    def get_team(
-        self,
-        team_id: str,
-    ):
-        """
-        Retrieves a specific team given its unique ID.
-
-        :param team_id: Unique team ID
-        :return: JSON response object
-        """
-        return self.agent.get(
-            f"{self.url}/v2/team/{team_id}",
-        ).json()
-
-    def patch_team(
-        self,
-        team_id: str,
-        name: str | None = None,
-        description: str | None = None,
-    ):
-        """
-        Updates a team given its unique ID.
-
-        :param team_id: Unique team ID
-        :param name: Name of team, defaults to None
-        :param description: Description of team, defaults to None
-        :return: JSON response object
-        """
-        filtered_json = filter_none({"name": name, "description": description})
-
-        return self.agent.patch(
-            f"{self.url}/v2/team/{team_id}",
-            json=filtered_json,
-        ).json()
-
     def get_access_request(self, model_id: str, access_request_id: str):
-        """
-        Retrieves a specific access request given its unique ID.
+        """Retrieve a specific access request given its unique ID.
 
         :param model_id: Unique model ID
         :param access_request_id: Unique access request ID
@@ -601,8 +566,7 @@ class Client:
         self,
         model_id: str,
     ):
-        """
-        Retrieves all access requests given a specific model.
+        """Retrieve all access requests given a specific model.
 
         :param model_id: Unique model ID
         :param access_request_id: Unique access request ID
@@ -613,8 +577,7 @@ class Client:
         ).json()
 
     def post_access_request(self, model_id: str, metadata: Any, schema_id: str):
-        """
-        Creates an access request given a model ID
+        """Create an access request given a model ID.
 
         :param model_id: Unique model ID
         :param metadata: Metadata object, defined by access request schema
@@ -627,8 +590,7 @@ class Client:
         ).json()
 
     def delete_access_request(self, model_id: str, access_request_id: str):
-        """
-        Deletes a specific access request associated with a model.
+        """Delete a specific access request associated with a model.
 
         :param model_id: Unique model ID
         :param access_request_id: Unique access request ID
@@ -645,8 +607,7 @@ class Client:
         metadata: Any,
         schema_id: str | None = None,
     ):
-        """
-        Updates an access request given its unique ID
+        """Update an access request given its unique ID.
 
         :param model_id: Unique model ID
         :param access_request_id: Unique access request ID
@@ -658,3 +619,41 @@ class Client:
             f"{self.url}/v2/model/{model_id}/access-request/{access_request_id}",
             json=filtered_json,
         ).json()
+
+    def put_file_scan(
+        self,
+        model_id: str,
+        file_id: str,
+    ):
+        """
+        Manually re-request a new antivirus scan for a file.
+
+        :param model_id: Unique model ID
+        :param file_id: Unique file ID
+        :return: JSON response object
+        """
+        return self.agent.put(
+            f"{self.url}/v2/filescanning/model/{model_id}/file/{file_id}/scan",
+        ).json()
+
+    def post_access_request_review(
+        self,
+        model_id: str,
+        access_request_id: str,
+        role: str,
+        decision: str,
+        comment: str | None = None,
+    ):
+        """Create a review for a release.
+
+        :param model_id: A unique model ID
+        :param access_request_id: Unique access request ID
+        :param role: The role of the user making the review
+        :param decision: Either approve or request changes
+        :param comment: A comment to go with the review
+        """
+        filtered_json = filter_none({"role": role, "decision": decision, "comment": comment})
+        return self.agent.post(
+            f"{self.url}/v2/model/{model_id}/access-request/{access_request_id}/review",
+            json=filtered_json,
+        ).json()
